/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.pc;

import java.util.regex.Pattern;


import javassist.CtClass;
import javassist.NotFoundException;
import javassist.compiler.CompileError;
import javassist.compiler.MemberResolver;
import javassist.gluonj.WeaveException;

public class PcPattern {
    private MemberResolver resolver;
    boolean not, subclass;
    private String className;
    private String memberName;
    private String descriptor;
    private Pattern classPat, memberPat;

    public PcPattern(MemberResolver mr, boolean notPat, String clazz,
                     boolean includesSubclass, String member, String desc)
    {
        resolver = mr;
        not = notPat;
        subclass = includesSubclass;
        className = clazz;
        memberName = member;
        descriptor = desc;
        classPat = clazz == null ? null : Pattern.compile(clazz);
        memberPat = member == null ? null : Pattern.compile(member);
    }

    public boolean isClassName() {
        return memberName == null && descriptor == null && className != null;
    }

    public boolean isMethod() {
        return memberName != null && descriptor != null;
    }

    public boolean isField() {
        return memberName != null && descriptor == null;
    }

    public boolean matchClass(String name) throws WeaveException {
        boolean yes = matchClass0(name);
        return not ? !yes : yes;
    }

    private boolean matchClass0(String name) throws WeaveException {
        if (name != null)
            if (match(classPat, name))
                return true;
            else
                if (subclass) {
                    try {
                        CtClass clazz = resolver.lookupClass(name, false);
                        if (matchClass0(clazz))
                            return true;
                    }
                    catch (CompileError ce) {
                        throw new WeaveException(ce);
                    }
                    catch (NotFoundException ne) {
                        throw new WeaveException(ne);
                    }
                }

        return false;
    }

    private boolean matchClass0(CtClass clazz) throws NotFoundException {
        if (clazz == null)
            return false;

        if (match(classPat, clazz.getName()))
            return true;

        CtClass parent = clazz.getSuperclass();
            if (matchClass0(parent))
                return true;

        CtClass[] intfs = clazz.getInterfaces();
        for (int i = 0; i < intfs.length; i++)
            if (matchClass0(intfs[0]))
                return true;

        return false;
    }

    private boolean match(Pattern pat, String name) {
        return pat.matcher(name).matches();
    }

    public boolean matchField(String clazz, String field) throws WeaveException {
        boolean yes = (className == null || matchClass0(clazz))
                      && match(memberPat, field);
        return not ? !yes : yes;   
    }

    public boolean matchMethod(String declClass, String name, String desc) throws WeaveException {
        boolean yes = (className == null || matchClass0(declClass))
                      && match(memberPat, name)
                      && desc != null && desc.startsWith(descriptor);
        return not ? !yes : yes;
    }
}
