/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.LoaderClassPath;
import javassist.NotFoundException;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;

import javassist.gluonj.Gluon;
import javassist.gluonj.WeaveException;

/**
 * Load-time weaver.  It creates an instance of a subclass of Weaver.
 * That instance is stored in the weaver field in HotSwapper.Transformer.
 *
 * @see Weaver
 */
public class HotSwapper {
    /**
     * Main method.
     *
     * <p>The argument is the name of a glue class.  If the weaver runs in
     * the debug mode, the argument is "&lt;<i>glue name</i>&gt;,debug".
     */
    public static void premain(String args, Instrumentation inst)
        throws WeaveException
    {
        args = parseArgs(args);
        if (args == null || args.length() < 1) {
            fatalError("no glue specified");    // never returns.
            // throw new WeaveException("no glue specified");
        }

        inst.addTransformer(new Transformer(args));
    }

    private static String parseArgs(String args) {
        if (args == null)
            return null;

        int end = args.indexOf(',');
        if (end >= 0) {
            String orig = args;
            args = args.substring(0, end);
            Gluon.stackTrace
                = orig.substring(end).equals(",debug");
        }

        return args;
    }

    /**
     * An event handler invoked when a new class file is being loaded.
     */
    public static class Transformer implements ClassFileTransformer {
        private String glueName;
        private ClassPool pool;
        private Weaver weaver;
        private boolean stop;

        public Transformer(String glue) {
            glueName = glue;
            pool = null;
            weaver = null;
            stop = false;
        }

        private void setupWeaver(ClassLoader cl) throws WeaveException {
            if (weaver != null)
                return;

            ClassPool cp = new ClassPool(true);
            if (cl != null)
                cp.insertClassPath(new LoaderClassPath(cl));

            pool = cp;
            Gluon g = Gluon.make(glueName, cp);
            weaver = new Weaver(g) {
                protected void save(CtClass clazz, boolean modify)
                    throws CannotCompileException, IOException {}
            };
        }

        public byte[] transform(ClassLoader loader, String className, Class classBeingRedefined,
                                ProtectionDomain domain, byte[] classfile)
            throws IllegalClassFormatException
        {
            if (stop || isNonTransformable(className))
                return null;

            try {
                setupWeaver(loader);
            }
            catch (Throwable t) {
                stop = true;
                showError("cannot read a glue: " + glueName, t);
                return null;
            }

            try {
                setupWeaver(loader);
                CtClass clazz = makeClass(className, classfile);
                if (weaver.transform(clazz.getName()))
                    return clazz.toBytecode();
                else
                    return classfile;
            }
            catch (Throwable t) {
                showTransformError(className, t);
                return null;
            }
        }

        private CtClass makeClass(String className, byte[] classfile)
            throws IOException, WeaveException
        {
            int counter = 0;
            do {
                try {
                    return pool.makeClass(new ByteArrayInputStream(classfile));
                }
                catch (RuntimeException re) {}   // the class is already transformed.
                try {
                    pool.get(className.replace('/', '.')).detach();
                }
                catch (NotFoundException e) {
                    counter = 1;
                }
            } while (counter++ == 0);
            throw new WeaveException("cannot reload (broken class file?): "
                                     + className.replace('/', '.'));
        }

        public boolean isNonTransformable(String jvmName) {
            return jvmName.startsWith("java/")
                || jvmName.startsWith("javax/")
                || jvmName.startsWith("com/sun/")
                || jvmName.startsWith("sun/")
                || jvmName.startsWith("sunw/");
                // || jvmName.startsWith("javassist/");
        }

        private void showTransformError(String className, Throwable e) {
            String msg = "cannot transform a class: " + className.replace('/', '.');
            showError(msg, e);
        }

        private void showError(String msg, Throwable e) {
            System.err.println("GluonJ Error: " + msg);
            System.err.println("  by " + e);
            if (Gluon.stackTrace)
                e.printStackTrace(System.err);
        }
    }

    private static void fatalError(String msg) {
        System.err.println("GluonJ Fatal Error: " + msg);
        System.err.println("** exit **");
        System.exit(1);
    }
}
