package javassist.gluonj.weave;

import javassist.gluonj.WeaveException;
import javassist.gluonj.pc.AndPc;
import javassist.gluonj.pc.CallPc;
import javassist.gluonj.pc.CflowPc;
import javassist.gluonj.pc.GetPc;
import javassist.gluonj.pc.IfPc;
import javassist.gluonj.pc.OrPc;
import javassist.gluonj.pc.NotPc;
import javassist.gluonj.pc.PointcutVisitor;
import javassist.gluonj.pc.SetPc;
import javassist.gluonj.pc.WithinPc;
import javassist.gluonj.util.SimplePcNode;

/**
 * Matcher is created for each join point and used as a visitor while
 * a pointcut expression is evaluated.
 */
public abstract class Matcher implements PointcutVisitor {
    /**
     * visit() sets this variable to true if a given pointcut expression
     * matches the join point represented by this matcher.
     */
    protected boolean result = false;

    /**
     * visit() may store a residue (a conditional expression evaluated at
     * runtime to select a join point) in this variable.
     */
    protected Residue residue = null;

    public boolean getResult() { return result; }

    public Residue getResidue() { return residue; }

    public void visit(AndPc node) throws WeaveException {
        residue = null;
        node.first().accept(this);
        if (result) {
            AndPc right = node.rest();
            if (right == null)
                return;

            Residue left = residue;
            residue = null;
            right.accept(this);
            if (result) {
                residue = Residue.and(left, residue);
                return;
            }
        }

        residue = null;
    }

    public void visit(OrPc node) throws WeaveException {
        residue = null;
        node.first().accept(this);
        boolean left = result;
        Residue leftR = residue;
        if (left && leftR == null)
            return;     // result == true

        OrPc right = node.rest();
        if (right == null)
            return;

        residue = null;
        right.accept(this);
        if (!left && !result) {
            residue = null;
            return;     // result == false
        }

        residue = Residue.or(leftR, residue);
        result = true;
    }

    public void visit(NotPc node) throws WeaveException {
        residue = null;
        node.getExpr().accept(this);
        if (result)
            if (residue == null)
                result = false;
            else
                residue = Residue.not(residue);
        else
            result = true;
    }

    public void visit(IfPc pc) throws WeaveException {
        residue = new Residue(pc.getExpr());
        result = true;
    }

    public void visit(CflowPc pc) throws WeaveException {
        residue = new Residue(pc.getExpr());
        result = true;
    }

    public void visit(WithinPc pc) throws WeaveException {
        result = false;
    }

    public void visit(CallPc pc) throws WeaveException {
        result = false;
    }

    public void visit(GetPc pc) throws WeaveException {
        result = false;
    }

    public void visit(SetPc pc) throws WeaveException {
        result = false;
    }

    public void visit(SimplePcNode pc) throws WeaveException {
        result = false;
    }
}
