/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CannotCompileException;
import javassist.NotFoundException;
import javassist.Loader;
import javassist.gluonj.Glue;
import javassist.gluonj.Gluon;
import javassist.gluonj.Refine;
import javassist.gluonj.WeaveException;

/**
 * Compile-time weaver.
 *
 * @see javassist.gluonj.ant.taksdefs.Weave
 */
public class OfflineWeaver extends Weaver {
    private String directory;
    private ArrayList notSavedClasses;

    /**
     * Constructs an offline weaver.
     * @throws WeaveException 
     */
    public OfflineWeaver(String glueName, String[] classpath)
        throws NotFoundException, WeaveException
    {
        super(makeGlue(glueName, classpath));
        directory = ".";
        notSavedClasses = new ArrayList();
    }

    private static Gluon makeGlue(String glueName, String[] classpath)
        throws NotFoundException, WeaveException
    {
        ClassPool cp = new ClassPool(true);
        ClassPool cp2 = new ClassPool(true);
        if (classpath != null)
            for (int i = 0; i < classpath.length; i++) {
                String path = dirName(classpath[i]);
                cp.appendClassPath(path);
                cp2.appendClassPath(path);
            }

        Loader cloader = new Loader(Gluon.class.getClassLoader(), cp);
        cloader.delegateLoadingOf("javassist.");
        return Gluon.make(glueName, cp2, cloader);
    }

    private static String dirName(String dir) {
        String path = dir.replace(File.separatorChar, '/');
        int len = path.length();
        if (path.charAt(len - 1) == '/')
            path = path.substring(0, len - 1);

        return path;
    }

    /**
     * Sets the output directory where class files are written.
     */
    public void setDirectory(String dir) {
        directory = dirName(dir);
    }

    protected void save(CtClass clazz, boolean modify)
        throws CannotCompileException, IOException
    {
        if (modify)
            clazz.writeFile(directory);
        else
            notSavedClasses.add(clazz);
    }

    private void saveAll()
        throws CannotCompileException, IOException
    {
        if (directory.equals("."))
            return;

        Iterator it = notSavedClasses.iterator();
        while (it.hasNext()) {
            CtClass cc = (CtClass)it.next();
            cc.writeFile(directory);
        }
    }

    /**
     * Transforms class files.  It may print an error message.
     */
    public void transformFiles(File baseDir, String[] fileNames) throws WeaveException {
        int len = fileNames.length;
        notSavedClasses.ensureCapacity(len);
        ArrayList classes = new ArrayList(len);     // ArrayList<CtClass>
        ClassPool cp = getClassPool();
        List inspectors = getGlue().getInspectors();
        for (int i = 0; i < len; i++) {
            String fname = fileNames[i];
            if (fname.endsWith(".class"))
                try {
                    readFile(cp, baseDir, fname, classes, inspectors);
                }
                catch (WeaveException e) {
                    showTransformError(fname, e);
                    throw e;
                }
        }

        Iterator it = classes.iterator();
        while (it.hasNext()) {
            CtClass c = (CtClass)it.next();
            try {
                /* if inspectors == null, then classes
                 * should be empty.  transform() has been
                 * already invoked.  See readFile().
                 */
                transform(c.getName(), false);
            }
            catch (WeaveException e) {
                showTransformError(c.getName(), e);
                throw e;
            }
        }

        Exception ex;
        try {
            saveAll();
            return;
        }
        catch (IOException e) { ex = e; }
        catch (CannotCompileException e) { ex = e; }
        printError(ex.getMessage());
        throw new WeaveException(ex); 
    }

    private void readFile(ClassPool cp, File baseDir, String name,
                          ArrayList classes, List inspectors)
        throws WeaveException
    {
        try {
            InputStream is
                = new BufferedInputStream(new FileInputStream(new File(baseDir, name)));
            CtClass clazz = cp.makeClass(is);
            is.close();
            if (notGlue(clazz)) {
                if (inspectors != null) {
                    runInspectors(clazz, inspectors);
                    classes.add(clazz);
                }
                else
                    transform(clazz.getName(), false);
            }
        }
        catch (IOException e) {
            throw new WeaveException(e);
        }
    }

    private boolean notGlue(CtClass clazz) {
        boolean result = true;
        Object[] anno;
        try {
            anno = clazz.getAnnotations();
        }
        catch (ClassNotFoundException e) {
            return true;
        }

        for (int j = 0; j < anno.length; j++)
            if (anno[j] instanceof Glue || anno[j] instanceof Refine)
                result = false;

        return result;
    }

    public static void main(String[] args) throws Exception {
        run0(OfflineWeaver.class.getName(), args);
    }

    public static void run(String mainClass, String[] args) {
        try {
            run0(mainClass, args);
        }
        catch (Exception e) {
            if (Gluon.stackTrace)
                e.printStackTrace(System.err);
        }
    }

    private static void run0(String mainClass, String[] args) throws Exception {
        if (args.length < 2) {
            showUsage(mainClass);
            return;
        }

        String directory = ".";
        int index = 0;
        if (args[0].equals("-d")) {
            if (args.length < 4) {
                showUsage(mainClass);
                return;
            }

            directory = args[1];
            index = 2;
        }

        String[] srcDirs = { "." };
        OfflineWeaver w = make(args[index], srcDirs);
        w.setDirectory(directory);
        String[] fileNames = new String[args.length - index - 1];
        for (int i = index + 1; i < args.length; i++)
            fileNames[i - index - 1] = args[i];

        w.transformFiles(null, fileNames);
        w.done();
    }

    public static OfflineWeaver make(String glueName, String[] classpath)
        throws NotFoundException, WeaveException
    {
        try {
            return new OfflineWeaver(glueName, classpath);
        }
        catch (NotFoundException e) {
            printError("cannot find a glue " + glueName);
            throw e;
        }
        catch (WeaveException e) {
            printError("cannot read a glue " + glueName);
            throw e;
        }
    }

    private static void showUsage(String mainClass) {
        System.err.println("Usage: java "
                + mainClass + " [ -d <output direcotry> ]"
                + " <glue class name> <java class file> ...");
    }

    public static void showTransformError(String className, WeaveException e) {
        String msg = "cannot transform a class: " + className.replace('/', '.');
        printError(msg);
        System.err.println("  by " + e);
    }

    private static void printError(String msg) {
        System.err.println("GluonJ Error: " + msg);
    }
}
