/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.util.HashMap;
import javassist.*;
import javassist.bytecode.ConstantAttribute;
import javassist.bytecode.FieldInfo;
import javassist.expr.ExprEditor;
import javassist.expr.MethodCall;
import javassist.gluonj.WeaveException;
import javassist.gluonj.Abstract;

/**
 * Refine class processor.
 */
public class RefineWeaver {
    /**
     * Appends the interfaces and methods declared in the
     * mixin class to the original class.
     *
     * @param orig      the original class.
     * @param mixin     the mixin class.
     * @param mixinClassNames      null or a mapping between mixin class names and
     *                             original class names.
     * @return the modified original class.
     */
    public CtClass weave(CtClass orig, CtClass mixin, ClassMap mixinClassNames)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        boolean isIntf = orig.isInterface();
        if (isIntf != mixin.isInterface())
            throw new WeaveException("class/interface mismatch between "
                        + orig.getName() + " and " + mixin.getName());

        CtClass[] interfaces = mixin.getInterfaces();
        for (int i = 0; i < interfaces.length; i++)
            if (!orig.subtypeOf(interfaces[i]))
                orig.addInterface(interfaces[i]);

        copyFields(orig, mixin);
        if (!isIntf)
            copyConstructors(orig, mixin);

        copyClassInitializer(orig, mixin);
        copyMethods(orig, mixin, isIntf, mixinClassNames);

        return orig;
    }

    private void copyFields(CtClass orig, CtClass mixin)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtField[] newFields = mixin.getDeclaredFields();
        for (int i = 0; i < newFields.length; i++) {
            CtField f = newFields[i];
            CtField origF = null;
            boolean inSuperClass = false;
            try {
                if (Modifier.isFinal(f.getModifiers()))
                    throw new WeaveException("cannot add/substitute a final field", f);

                try {
                    origF = orig.getDeclaredField(f.getName());
                }
                catch (NotFoundException e) {
                    // may throw a NotFoundException
                    origF = orig.getField(f.getName());
                    inSuperClass = true;
                    if (!origF.visibleFrom(orig))
                        throw new WeaveException(orig.getName() + " cannot access a field", origF);
                }

                if (!origF.getSignature().equals(f.getSignature()))
                    throw new WeaveException("incompatible field type", f);

                if (Modifier.isFinal(origF.getModifiers()) && hasInitializer(f))
                    throw new WeaveException("cannot update a final field", origF);
            }
            catch (NotFoundException e) {}
            if (origF == null || hasInitializer(f)) {
                if (inSuperClass)
                    throw new WeaveException("cannot override a super-class field", origF);

                CtField newF = new CtField(f, orig);    // ConstantAttribute is not copied.
                if (origF != null)
                    orig.removeField(origF);

                orig.addField(newF);
            }
        }
    }

    private static boolean hasInitializer(CtField f) {
        FieldInfo info = f.getFieldInfo2();
        return info.getAttribute(ConstantAttribute.tag) != null;
    }

    private static void copyConstructors(CtClass orig, CtClass mixin)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtConstructor[] cons = mixin.getDeclaredConstructors();
        if (cons.length != 1)
            throw new WeaveException("more than one constructors in "
                                     + mixin.getName());

        CtConstructor cc = cons[0];
        if (cc.isEmpty())       // defualt constructor
            return;

        if (!cc.getSignature().equals("()V"))
            throw new WeaveException("constructor with parameters in " + mixin.getName());

        String methodName = orig.makeUniqueName("_init");
        CtMethod mth = cc.toMethod(methodName, orig);
        mth.setModifiers(Modifier.PRIVATE);
        orig.addMethod(mth);
        cons = orig.getDeclaredConstructors();
        String expr = methodName + "();";
        for (int i = 0; i < cons.length; i++) {
            CtConstructor c = cons[i];
            if (c.callsSuper())
                c.insertAfter(expr);
        }
    }

    private static void copyClassInitializer(CtClass orig, CtClass mixin)
        throws WeaveException, CannotCompileException, NotFoundException
    {
        CtConstructor cons = mixin.getClassInitializer();
        if (cons == null || cons.isEmpty())
            return;

        CtConstructor dest = orig.getClassInitializer();
        if (dest == null) {
            dest = new CtConstructor(cons, orig, null);
            orig.addConstructor(dest);
        }
        else {
            String methodName = orig.makeUniqueName("_clinit");
            CtMethod mth = cons.toMethod(methodName, orig);
            mth.setModifiers(Modifier.PRIVATE | Modifier.STATIC);
            orig.addMethod(mth);
            dest.insertAfter(methodName + "();");
        }
    }

    private void copyMethods(CtClass orig, CtClass mixin, boolean isIntf, ClassMap mixinNames)
        throws WeaveException, NotFoundException, CannotCompileException
    {
        HashMap hash = new HashMap();
        CtMethod[] origMethods = orig.getDeclaredMethods();
        for (int i = 0; i < origMethods.length; i++) {
            CtMethod m = origMethods[i];
            hash.put(m, m);
        }

        HashMap allHash = new HashMap();
        CtMethod[] allOrigMethods = orig.getMethods();
        for (int i = 0; i < allOrigMethods.length; i++) {
            CtMethod m = allOrigMethods[i];
            allHash.put(m, m);
        }

        CtMethod[] newMethods = mixin.getDeclaredMethods();
        for (int i = 0; i < newMethods.length; i++) {
            CtMethod m = newMethods[i];
            CtMethod found = (CtMethod)hash.get(m);
            if (skipAbstractMethod(orig, isIntf, m, found, allHash))
                continue;

            CtMethod newM = CtNewMethod.copy(m, orig, mixinNames);
            if (found == null) {
                if (hasOrigCall(m))
                    throw new WeaveException("no original method of "
                                             + m.getName());
            }
            else
                if (hasOrigCall(m)) {
                    renameOrigMethod(orig, found, newM);
                }
                else
                    orig.removeMethod(found);

            orig.addMethod(newM);
        }
    }

    private boolean skipAbstractMethod(CtClass orig, boolean isIfs, CtMethod m,
                                       CtMethod found, HashMap allHash)
        throws WeaveException
    {
        if (Modifier.isAbstract(m.getModifiers()) || isAbstract(m)) {
            if (isIfs) {    // is an interface type
                if (found == null)
                    found = (CtMethod)allHash.get(m);

                return found != null;
            }

            if (found == null && !m.getName().startsWith(origPrefix)) {
                found = (CtMethod)allHash.get(m);
                if (found == null || !isVisible(orig, found))
                    throw new WeaveException(orig.getName()
                            + " cannot find or access "
                            + m.getName() + m.getSignature());
            }

            return true;
        }
        else
            return false;
    }

    /*
     * Returns true if this (static) method is annotated with @Abstract.
     */
    private boolean isAbstract(CtMethod m) {
        try {
            Object[] anno = m.getAnnotations();
            int n = anno.length;
            for (int i = 0; i < n; i++)
                if (anno[i] instanceof Abstract)
                    return true;
        }
        catch (ClassNotFoundException e) {}
        return false;
    }

    private void renameOrigMethod(CtClass orig, CtMethod origM, CtMethod newM) throws CannotCompileException {
        String methodName = origPrefix + origM.getName();
        CtMethod conflict;
        try {
            conflict = orig.getMethod(methodName, origM.getSignature());
        }
        catch (NotFoundException e) {
            origM.setName(methodName);
            return;
        }

        /* origM must be renamed to another name.
         * methodName is already used.
         */
        String anotherName = orig.makeUniqueName(methodName);
        origM.setName(anotherName);
        CodeConverter conv = new CodeConverter();
        conv.redirectMethodCall(conflict, origM);
        newM.instrument(conv);
    }

    /**
     * Returns true if the method is visible from the class. 
     */
    private static boolean isVisible(CtClass from, CtMethod meth) {
        if (!Modifier.isPackage(meth.getModifiers()))
            return true;
        else {
            String p = from.getPackageName();
            String q = meth.getDeclaringClass().getPackageName();
            if (p == null)
                return q == null;
            else
                return p.equals(q);
        }
    }

    private static final String origPrefix = "orig_";

    private boolean hasOrigCall(CtMethod method)
        throws WeaveException
    {
        final boolean[] found = { false };
        final String methodName = method.getName();
        final String methodSig = method.getSignature();
        try {
            method.instrument(new ExprEditor() {
                public void edit(MethodCall call)
                    throws CannotCompileException
                {
                    String name = call.getMethodName();
                    if (name.startsWith(origPrefix))
                        if (name.equals(origPrefix + methodName)
                                && call.getSignature().equals(methodSig))
                            found[0] = true;
                        else
                            throw new CannotCompileException("cannot call "
                                    + name + call.getSignature());
                }
            });
        }
        catch (CannotCompileException e) {
            throw new WeaveException(e);
        }

        return found[0];
    }
}
